@@ -21,20 +21,23 @@
  */
 #include "dnsdist.hh"
 #include "dnsdist-ecs.hh"
+#include "dnsdist-rings.hh"
+
 #include "dnsparser.hh"
 #include "ednsoptions.hh"
 #include "dolog.hh"
 #include "lock.hh"
 #include "gettime.hh"
+#include "tcpiohandler.hh"
 #include <thread>
 #include <atomic>
 
 using std::thread;
 using std::atomic;
 
-/* TCP: the grand design. 
-   We forward 'messages' between clients and downstream servers. Messages are 65k bytes large, tops. 
-   An answer might theoretically consist of multiple messages (for example, in the case of AXFR), initially 
+/* TCP: the grand design.
+   We forward 'messages' between clients and downstream servers. Messages are 65k bytes large, tops.
+   An answer might theoretically consist of multiple messages (for example, in the case of AXFR), initially
    we will not go there.
 
    In a sense there is a strong symmetry between UDP and TCP, once a connection to a downstream has been setup.
@@ -184,15 +187,28 @@ catch(...) {
   return false;
 }
 
-static bool sendResponseToClient(int fd, const char* response, uint16_t responseLen)
+static bool getNonBlockingMsgLenFromClient(TCPIOHandler& handler, uint16_t* len)
+try
 {
-  return sendSizeAndMsgWithTimeout(fd, responseLen, response, g_tcpSendTimeout, nullptr, nullptr, 0, 0, 0);
+  uint16_t raw;
+  size_t ret = handler.read(&raw, sizeof raw, g_tcpRecvTimeout);
+  if(ret != sizeof raw)
+    return false;
+  *len = ntohs(raw);
+  return true;
+}
+catch(...) {
+  return false;
 }
 
 static bool maxConnectionDurationReached(unsigned int maxConnectionDuration, time_t start, unsigned int& remainingTime)
 {
   if (maxConnectionDuration) {
-    time_t elapsed = time(NULL) - start;
+    time_t curtime = time(nullptr);
+    unsigned int elapsed = 0;
+    if (curtime > start) { // To prevent issues when time goes backward
+      elapsed = curtime - start;
+    }
     if (elapsed >= maxConnectionDuration) {
       return true;
     }
@@ -220,21 +236,12 @@ void* tcpClientThread(int pipefd)
 {
   /* we get launched with a pipe on which we receive file descriptors from clients that we own
      from that point on */
-     
+
   bool outstanding = false;
   time_t lastTCPCleanup = time(nullptr);
   
-     
-  auto localPolicy = g_policy.getLocal();
-  auto localRulactions = g_rulactions.getLocal();
+  LocalHolders holders;
   auto localRespRulactions = g_resprulactions.getLocal();
-  auto localCacheHitRespRulactions = g_cachehitresprulactions.getLocal();
-  auto localDynBlockNMG = g_dynblockNMG.getLocal();
-  auto localDynBlockSMT = g_dynblockSMT.getLocal();
-  auto localPools = g_pools.getLocal();
-#ifdef HAVE_PROTOBUF
-  boost::uuids::random_generator uuidGenerator;
-#endif
 #ifdef HAVE_DNSCRYPT
   /* when the answer is encrypted in place, we need to get a copy
      of the original header before encryption to fill the ring buffer */
@@ -254,41 +261,46 @@ void* tcpClientThread(int pipefd)
 
     g_tcpclientthreads->decrementQueuedCount();
     ci=*citmp;
-    delete citmp;    
+    delete citmp;
 
     uint16_t qlen, rlen;
-    string largerQuery;
     vector<uint8_t> rewrittenResponse;
     shared_ptr<DownstreamState> ds;
     ComboAddress dest;
-    memset(&dest, 0, sizeof(dest));
+    dest.reset();
     dest.sin4.sin_family = ci.remote.sin4.sin_family;
     socklen_t len = dest.getSocklen();
     size_t queriesCount = 0;
     time_t connectionStartTime = time(NULL);
-
-    if (!setNonBlocking(ci.fd))
-      goto drop;
+    std::vector<char> queryBuffer;
+    std::vector<char> answerBuffer;
 
     if (getsockname(ci.fd, (sockaddr*)&dest, &len)) {
       dest = ci.cs->local;
     }
 
     try {
+      TCPIOHandler handler(ci.fd, g_tcpRecvTimeout, ci.cs->tlsFrontend ? ci.cs->tlsFrontend->getContext() : nullptr, connectionStartTime);
+
       for(;;) {
         unsigned int remainingTime = 0;
         ds = nullptr;
         outstanding = false;
 
-        if(!getNonBlockingMsgLen(ci.fd, &qlen, g_tcpRecvTimeout)) {
+        if(!getNonBlockingMsgLenFromClient(handler, &qlen)) {
+          break;
+        }
+
+        queriesCount++;
+
+        if (qlen < sizeof(dnsheader)) {
+          g_stats.nonCompliantQueries++;
           break;
         }
 
         ci.cs->queries++;
         g_stats.queries++;
 
-        queriesCount++;
-
         if (g_maxTCPQueriesPerConn && queriesCount > g_maxTCPQueriesPerConn) {
           vinfolog("Terminating TCP connection from %s because it reached the maximum number of queries per conn (%d / %d)", ci.remote.toStringWithPort(), queriesCount, g_maxTCPQueriesPerConn);
           break;
@@ -299,124 +311,121 @@ void* tcpClientThread(int pipefd)
           break;
         }
 
-        if (qlen < sizeof(dnsheader)) {
-          g_stats.nonCompliantQueries++;
-          break;
-        }
-
         bool ednsAdded = false;
         bool ecsAdded = false;
-        /* if the query is small, allocate a bit more
-           memory to be able to spoof the content,
+        /* allocate a bit more memory to be able to spoof the content,
            or to add ECS without allocating a new buffer */
-        size_t querySize = qlen <= 4096 ? qlen + 512 : qlen;
-        char queryBuffer[querySize];
-        const char* query = queryBuffer;
-        readn2WithTimeout(ci.fd, queryBuffer, qlen, g_tcpRecvTimeout, remainingTime);
+        queryBuffer.resize(qlen + 512);
+
+        char* query = &queryBuffer[0];
+        handler.read(query, qlen, g_tcpRecvTimeout, remainingTime);
+
+        /* we need this one to be accurate ("real") for the protobuf message */
+	struct timespec queryRealTime;
+	struct timespec now;
+	gettime(&now);
+	gettime(&queryRealTime, true);
 
 #ifdef HAVE_DNSCRYPT
-        std::shared_ptr<DnsCryptQuery> dnsCryptQuery = 0;
+        std::shared_ptr<DNSCryptQuery> dnsCryptQuery = nullptr;
 
         if (ci.cs->dnscryptCtx) {
-          dnsCryptQuery = std::make_shared<DnsCryptQuery>();
+          dnsCryptQuery = std::make_shared<DNSCryptQuery>(ci.cs->dnscryptCtx);
           uint16_t decryptedQueryLen = 0;
           vector<uint8_t> response;
-          bool decrypted = handleDnsCryptQuery(ci.cs->dnscryptCtx, queryBuffer, qlen, dnsCryptQuery, &decryptedQueryLen, true, response);
+          bool decrypted = handleDNSCryptQuery(query, qlen, dnsCryptQuery, &decryptedQueryLen, true, queryRealTime.tv_sec, response);
 
           if (!decrypted) {
             if (response.size() > 0) {
-              sendResponseToClient(ci.fd, reinterpret_cast<char*>(response.data()), (uint16_t) response.size());
+              handler.writeSizeAndMsg(response.data(), response.size(), g_tcpSendTimeout);
             }
             break;
           }
           qlen = decryptedQueryLen;
         }
 #endif
-        struct dnsheader* dh = (struct dnsheader*) query;
+        struct dnsheader* dh = reinterpret_cast<struct dnsheader*>(query);
 
-        if(dh->qr) {   // don't respond to responses
-          g_stats.nonCompliantQueries++;
+        if (!checkQueryHeaders(dh)) {
           goto drop;
         }
 
-        if(dh->qdcount == 0) {
-          g_stats.emptyQueries++;
-          goto drop;
-        }
-
-        if (dh->rd) {
-          g_stats.rdQueries++;
-        }
+	string poolname;
+	int delayMsec=0;
 
 	const uint16_t* flags = getFlagsFromDNSHeader(dh);
 	uint16_t origFlags = *flags;
 	uint16_t qtype, qclass;
 	unsigned int consumed = 0;
 	DNSName qname(query, qlen, sizeof(dnsheader), false, &qtype, &qclass, &consumed);
-	DNSQuestion dq(&qname, qtype, qclass, &dest, &ci.remote, (dnsheader*)query, querySize, qlen, true);
-#ifdef HAVE_PROTOBUF
-        dq.uniqueId = uuidGenerator();
-#endif
+	DNSQuestion dq(&qname, qtype, qclass, &dest, &ci.remote, dh, queryBuffer.size(), qlen, true, &queryRealTime);
 
-	string poolname;
-	int delayMsec=0;
-	/* we need this one to be accurate ("real") for the protobuf message */
-	struct timespec queryRealTime;
-	struct timespec now;
-	gettime(&now);
-	gettime(&queryRealTime, true);
-
-	if (!processQuery(localDynBlockNMG, localDynBlockSMT, localRulactions, dq, poolname, &delayMsec, now)) {
+	if (!processQuery(holders, dq, poolname, &delayMsec, now)) {
 	  goto drop;
 	}
 
 	if(dq.dh->qr) { // something turned it into a response
           restoreFlags(dh, origFlags);
+
+          DNSResponse dr(dq.qname, dq.qtype, dq.qclass, dq.local, dq.remote, reinterpret_cast<dnsheader*>(query), dq.size, dq.len, true, &queryRealTime);
+#ifdef HAVE_PROTOBUF
+          dr.uniqueId = dq.uniqueId;
+#endif
+          dr.qTag = dq.qTag;
+
+          if (!processResponse(holders.selfAnsweredRespRulactions, dr, &delayMsec)) {
+            goto drop;
+          }
+
 #ifdef HAVE_DNSCRYPT
-          if (!encryptResponse(queryBuffer, &dq.len, dq.size, true, dnsCryptQuery, nullptr, nullptr)) {
+          if (!encryptResponse(query, &dq.len, dq.size, true, dnsCryptQuery, nullptr, nullptr)) {
             goto drop;
           }
 #endif
-          sendResponseToClient(ci.fd, query, dq.len);
-	  g_stats.selfAnswered++;
-	  continue;
-	}
+          handler.writeSizeAndMsg(query, dq.len, g_tcpSendTimeout);
+          g_stats.selfAnswered++;
+          continue;
+        }
 
-        std::shared_ptr<ServerPool> serverPool = getPool(*localPools, poolname);
-        std::shared_ptr<DNSDistPacketCache> packetCache = nullptr;
-        auto policy = localPolicy->policy;
+        std::shared_ptr<ServerPool> serverPool = getPool(*holders.pools, poolname);
+        std::shared_ptr<DNSDistPacketCache> packetCache = serverPool->packetCache;
+
+        auto policy = *(holders.policy);
         if (serverPool->policy != nullptr) {
-          policy = serverPool->policy->policy;
+          policy = *(serverPool->policy);
         }
-        {
+        auto servers = serverPool->getServers();
+        if (policy.isLua) {
           std::lock_guard<std::mutex> lock(g_luamutex);
-          ds = policy(serverPool->servers, &dq);
-          packetCache = serverPool->packetCache;
+          ds = policy.policy(servers, &dq);
+        }
+        else {
+          ds = policy.policy(servers, &dq);
         }
 
-        if (dq.useECS && ds && ds->useECS) {
+        if (dq.useECS && ((ds && ds->useECS) || (!ds && serverPool->getECS()))) {
           uint16_t newLen = dq.len;
-          handleEDNSClientSubnet(queryBuffer, dq.size, consumed, &newLen, largerQuery, &ednsAdded, &ecsAdded, ci.remote, dq.ecsOverride, dq.ecsPrefixLength);
-          if (largerQuery.empty() == false) {
-            query = largerQuery.c_str();
-            dq.len = (uint16_t) largerQuery.size();
-            dq.size = largerQuery.size();
-          } else {
-            dq.len = newLen;
+          if (!handleEDNSClientSubnet(query, dq.size, consumed, &newLen, &ednsAdded, &ecsAdded, dq.ecsSet ? dq.ecs.getNetwork() : ci.remote, dq.ecsOverride, dq.ecsSet ? dq.ecs.getBits() : dq.ecsPrefixLength)) {
+            vinfolog("Dropping query from %s because we couldn't insert the ECS value", ci.remote.toStringWithPort());
+            goto drop;
           }
+          dq.len = newLen;
         }
 
         uint32_t cacheKey = 0;
+        boost::optional<Netmask> subnet;
         if (packetCache && !dq.skipCache) {
           char cachedResponse[4096];
           uint16_t cachedResponseSize = sizeof cachedResponse;
           uint32_t allowExpired = ds ? 0 : g_staleCacheEntriesTTL;
-          if (packetCache->get(dq, (uint16_t) consumed, dq.dh->id, cachedResponse, &cachedResponseSize, &cacheKey, allowExpired)) {
+          if (packetCache->get(dq, (uint16_t) consumed, dq.dh->id, cachedResponse, &cachedResponseSize, &cacheKey, subnet, allowExpired)) {
             DNSResponse dr(dq.qname, dq.qtype, dq.qclass, dq.local, dq.remote, (dnsheader*) cachedResponse, sizeof cachedResponse, cachedResponseSize, true, &queryRealTime);
 #ifdef HAVE_PROTOBUF
             dr.uniqueId = dq.uniqueId;
 #endif
-            if (!processResponse(localCacheHitRespRulactions, dr, &delayMsec)) {
+            dr.qTag = dq.qTag;
+
+            if (!processResponse(holders.cacheHitRespRulactions, dr, &delayMsec)) {
               goto drop;
             }
 
@@ -425,7 +434,7 @@ void* tcpClientThread(int pipefd)
               goto drop;
             }
 #endif
-            sendResponseToClient(ci.fd, cachedResponse, cachedResponseSize);
+            handler.writeSizeAndMsg(cachedResponse, cachedResponseSize, g_tcpSendTimeout);
             g_stats.cacheHits++;
             continue;
           }
@@ -440,18 +449,34 @@ void* tcpClientThread(int pipefd)
             dq.dh->rcode = RCode::ServFail;
             dq.dh->qr = true;
 
+            DNSResponse dr(dq.qname, dq.qtype, dq.qclass, dq.local, dq.remote, reinterpret_cast<dnsheader*>(query), dq.size, dq.len, false, &queryRealTime);
+#ifdef HAVE_PROTOBUF
+            dr.uniqueId = dq.uniqueId;
+#endif
+            dr.qTag = dq.qTag;
+
+            if (!processResponse(holders.selfAnsweredRespRulactions, dr, &delayMsec)) {
+              goto drop;
+            }
+
 #ifdef HAVE_DNSCRYPT
-            if (!encryptResponse(queryBuffer, &dq.len, dq.size, true, dnsCryptQuery, nullptr, nullptr)) {
+            if (!encryptResponse(query, &dq.len, dq.size, true, dnsCryptQuery, nullptr, nullptr)) {
               goto drop;
             }
 #endif
-            sendResponseToClient(ci.fd, query, dq.len);
+            handler.writeSizeAndMsg(query, dq.len, g_tcpSendTimeout);
+
+            // no response-only statistics counter to update.
             continue;
           }
 
           break;
         }
 
+        if (dq.addXPF && ds->xpfRRCode != 0) {
+          addXPF(dq, ds->xpfRRCode);
+        }
+
 	int dsock = -1;
 	uint16_t downstreamFailures=0;
 #ifdef MSG_FASTOPEN
@@ -496,7 +521,7 @@ void* tcpClientThread(int pipefd)
           sendSizeAndMsgWithTimeout(dsock, dq.len, query, ds->tcpSendTimeout, &ds->remote, &ds->sourceAddr, ds->sourceItf, 0, socketFlags);
         }
         catch(const runtime_error& e) {
-          vinfolog("Downstream connection to %s died on us, getting a new one!", ds->getName());
+          vinfolog("Downstream connection to %s died on us (%s), getting a new one!", ds->getName(), e.what());
           close(dsock);
           dsock=-1;
           sockets.erase(ds->remote);
@@ -514,7 +539,7 @@ void* tcpClientThread(int pipefd)
         if (isXFR) {
           dq.skipCache = true;
         }
-
+        bool firstPacket=true;
       getpacket:;
 
         if(!getNonBlockingMsgLen(dsock, &rlen, ds->tcpRecvTimeout)) {
@@ -542,9 +567,9 @@ void* tcpClientThread(int pipefd)
         }
 #endif
         responseSize += addRoom;
-        char answerbuffer[responseSize];
-        readn2WithTimeout(dsock, answerbuffer, rlen, ds->tcpRecvTimeout);
-        char* response = answerbuffer;
+        answerBuffer.resize(responseSize);
+        char* response = answerBuffer.data();
+        readn2WithTimeout(dsock, response, rlen, ds->tcpRecvTimeout);
         uint16_t responseLen = rlen;
         if (outstanding) {
           /* might be false for {A,I}XFR */
@@ -556,10 +581,10 @@ void* tcpClientThread(int pipefd)
           break;
         }
 
-        if (!responseContentMatches(response, responseLen, qname, qtype, qclass, ds->remote)) {
+        if (firstPacket && !responseContentMatches(response, responseLen, qname, qtype, qclass, ds->remote)) {
           break;
         }
-
+        firstPacket=false;
         if (!fixUpResponse(&response, &responseLen, &responseSize, qname, origFlags, ednsAdded, ecsAdded, rewrittenResponse, addRoom)) {
           break;
         }
@@ -569,12 +594,14 @@ void* tcpClientThread(int pipefd)
 #ifdef HAVE_PROTOBUF
         dr.uniqueId = dq.uniqueId;
 #endif
+        dr.qTag = dq.qTag;
+
         if (!processResponse(localRespRulactions, dr, &delayMsec)) {
           break;
         }
 
 	if (packetCache && !dq.skipCache) {
-	  packetCache->insert(cacheKey, qname, qtype, qclass, response, responseLen, true, dh->rcode);
+	  packetCache->insert(cacheKey, subnet, origFlags, qname, qtype, qclass, response, responseLen, true, dh->rcode, dq.tempFailureTTL);
 	}
 
 #ifdef HAVE_DNSCRYPT
@@ -582,7 +609,7 @@ void* tcpClientThread(int pipefd)
           goto drop;
         }
 #endif
-        if (!sendResponseToClient(ci.fd, response, responseLen)) {
+        if (!handler.writeSizeAndMsg(response, responseLen, g_tcpSendTimeout)) {
           break;
         }
 
@@ -608,24 +635,21 @@ void* tcpClientThread(int pipefd)
         struct timespec answertime;
         gettime(&answertime);
         unsigned int udiff = 1000000.0*DiffTime(now,answertime);
-        {
-          std::lock_guard<std::mutex> lock(g_rings.respMutex);
-          g_rings.respRing.push_back({answertime, ci.remote, qname, dq.qtype, (unsigned int)udiff, (unsigned int)responseLen, *dh, ds->remote});
-        }
+        g_rings.insertResponse(answertime, ci.remote, qname, dq.qtype, (unsigned int)udiff, (unsigned int)responseLen, *dh, ds->remote);
 
-        largerQuery.clear();
         rewrittenResponse.clear();
       }
     }
-    catch(...){}
+    catch(...) {}
 
   drop:;
-    
+
     vinfolog("Closing TCP client connection with %s", ci.remote.toStringWithPort());
     if (ci.fd >= 0) {
       close(ci.fd);
     }
     ci.fd = -1;
+
     if (ds && outstanding) {
       outstanding = false;
       --ds->outstanding;
@@ -658,10 +682,18 @@ void* tcpAcceptorThread(void* p)
     ConnectionInfo* ci = nullptr;
     tcpClientCountIncremented = false;
     try {
+      socklen_t remlen = remote.getSocklen();
       ci = new ConnectionInfo;
       ci->cs = cs;
       ci->fd = -1;
-      ci->fd = SAccept(cs->tcpFD, remote);
+#ifdef HAVE_ACCEPT4
+      ci->fd = accept4(cs->tcpFD, (struct sockaddr*)&remote, &remlen, SOCK_NONBLOCK);
+#else
+      ci->fd = accept(cs->tcpFD, (struct sockaddr*)&remote, &remlen);
+#endif
+      if(ci->fd < 0) {
+        throw std::runtime_error((boost::format("accepting new connection on socket: %s") % strerror(errno)).str());
+      }
 
       if(!acl->match(remote)) {
 	g_stats.aclDrops++;
@@ -672,6 +704,15 @@ void* tcpAcceptorThread(void* p)
 	continue;
       }
 
+#ifndef HAVE_ACCEPT4
+      if (!setNonBlocking(ci->fd)) {
+        close(ci->fd);
+        delete ci;
+        ci=nullptr;
+        continue;
+      }
+#endif
+
       if(g_maxTCPQueuedConnections > 0 && g_tcpclientthreads->getQueuedCount() >= g_maxTCPQueuedConnections) {
         close(ci->fd);
         delete ci;
@@ -732,7 +773,6 @@ void* tcpAcceptorThread(void* p)
   return 0;
 }
 
-
 bool getMsgLen32(int fd, uint32_t* len)
 try
 {
