@@ -57,7 +57,13 @@
 #include <sys/types.h>
 #include <pwd.h>
 #include <grp.h>
-
+#ifdef __FreeBSD__
+#  include <pthread_np.h>
+#endif
+#ifdef __NetBSD__
+#  include <pthread.h>
+#  include <sched.h>
+#endif
 
 bool g_singleThreaded;
 
@@ -332,10 +338,7 @@ int waitForRWData(int fd, bool waitForRead, int seconds, int useconds, bool* err
     pfd.events=POLLOUT;
 
   ret = poll(&pfd, 1, seconds * 1000 + useconds/1000);
-  if ( ret == -1 ) {
-    errno = ETIMEDOUT; // ???
-  }
-  else if (ret > 0) {
+  if (ret > 0) {
     if (error && (pfd.revents & POLLERR)) {
       *error = true;
     }
@@ -347,6 +350,44 @@ int waitForRWData(int fd, bool waitForRead, int seconds, int useconds, bool* err
   return ret;
 }
 
+// returns -1 in case of error, 0 if no data is available, 1 if there is. In the first two cases, errno is set
+int waitForMultiData(const set<int>& fds, const int seconds, const int useconds, int* fdOut) {
+  set<int> realFDs;
+  for (const auto& fd : fds) {
+    if (fd >= 0 && realFDs.count(fd) == 0) {
+      realFDs.insert(fd);
+    }
+  }
+
+  std::vector<struct pollfd> pfds(realFDs.size());
+  memset(pfds.data(), 0, realFDs.size()*sizeof(struct pollfd));
+  int ctr = 0;
+  for (const auto& fd : realFDs) {
+    pfds[ctr].fd = fd;
+    pfds[ctr].events = POLLIN;
+    ctr++;
+  }
+
+  int ret;
+  if(seconds >= 0)
+    ret = poll(pfds.data(), realFDs.size(), seconds * 1000 + useconds/1000);
+  else
+    ret = poll(pfds.data(), realFDs.size(), -1);
+  if(ret <= 0)
+    return ret;
+
+  set<int> pollinFDs;
+  for (const auto& pfd : pfds) {
+    if (pfd.revents & POLLIN) {
+      pollinFDs.insert(pfd.fd);
+    }
+  }
+  set<int>::const_iterator it(pollinFDs.begin());
+  advance(it, random() % pollinFDs.size());
+  *fdOut = *it;
+  return 1;
+}
+
 // returns -1 in case of error, 0 if no data is available, 1 if there is. In the first two cases, errno is set
 int waitFor2Data(int fd1, int fd2, int seconds, int useconds, int*fd)
 {
@@ -724,7 +765,7 @@ int makeIPv6sockaddr(const std::string& addr, struct sockaddr_in6* ret)
         port = pdns_stou(addr.substr(pos+2));
         portSet = true;
       }
-      catch(std::out_of_range) {
+      catch(const std::out_of_range&) {
         return -1;
       }
     }
@@ -864,7 +905,7 @@ void addCMsgSrcAddr(struct msghdr* msgh, void* cmsgbuf, const ComboAddress* sour
     pkt->ipi6_ifindex = itfIndex;
   }
   else {
-#ifdef IP_PKTINFO
+#if defined(IP_PKTINFO)
     struct in_pktinfo *pkt;
 
     msgh->msg_control = cmsgbuf;
@@ -879,8 +920,7 @@ void addCMsgSrcAddr(struct msghdr* msgh, void* cmsgbuf, const ComboAddress* sour
     memset(pkt, 0, sizeof(*pkt));
     pkt->ipi_spec_dst = source->sin4.sin_addr;
     pkt->ipi_ifindex = itfIndex;
-#endif
-#ifdef IP_SENDSRCADDR
+#elif defined(IP_SENDSRCADDR)
     struct in_addr *in;
 
     msgh->msg_control = cmsgbuf;
@@ -1015,8 +1055,9 @@ bool setSocketTimestamps(int fd)
 #ifdef SO_TIMESTAMP
   int on=1;
   return setsockopt(fd, SOL_SOCKET, SO_TIMESTAMP, (char*)&on, sizeof(on)) == 0;
-#endif
+#else
   return true; // we pretend this happened.
+#endif
 }
 
 bool setTCPNoDelay(int sock)
@@ -1046,6 +1087,14 @@ bool setBlocking(int sock)
   return true;
 }
 
+bool setReuseAddr(int sock)
+{
+  int tmp = 1;
+  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char*)&tmp, static_cast<unsigned>(sizeof tmp))<0)
+    throw PDNSException(string("Setsockopt failed: ")+strerror(errno));
+  return true;
+}
+
 bool isNonBlocking(int sock)
 {
   int flags=fcntl(sock,F_GETFL,0);
@@ -1308,9 +1357,18 @@ gid_t strToGID(const string &str)
 unsigned int pdns_stou(const std::string& str, size_t * idx, int base)
 {
   if (str.empty()) return 0; // compatibility
-  unsigned long result = std::stoul(str, idx, base);
+  unsigned long result;
+  try {
+    result = std::stoul(str, idx, base);
+  }
+  catch(std::invalid_argument& e) {
+    throw std::invalid_argument(string(e.what()) + "; (invalid argument during std::stoul); data was \""+str+"\"");
+  }
+  catch(std::out_of_range& e) {
+    throw std::out_of_range(string(e.what()) + "; (out of range during std::stoul); data was \""+str+"\"");
+  }
   if (result > std::numeric_limits<unsigned int>::max()) {
-    throw std::out_of_range("stou");
+    throw std::out_of_range("stoul returned result out of unsigned int range; data was \""+str+"\"");
   }
   return static_cast<unsigned int>(result);
 }
@@ -1327,6 +1385,20 @@ bool isSettingThreadCPUAffinitySupported()
 int mapThreadToCPUList(pthread_t tid, const std::set<int>& cpus)
 {
 #ifdef HAVE_PTHREAD_SETAFFINITY_NP
+#  ifdef __NetBSD__
+  cpuset_t *cpuset;
+  cpuset = cpuset_create();
+  for (const auto cpuID : cpus) {
+    cpuset_set(cpuID, cpuset);
+  }
+
+  return pthread_setaffinity_np(tid,
+                                cpuset_size(cpuset),
+                                cpuset);
+#  else
+#    ifdef __FreeBSD__
+#      define cpu_set_t cpuset_t
+#    endif
   cpu_set_t cpuset;
   CPU_ZERO(&cpuset);
   for (const auto cpuID : cpus) {
@@ -1336,6 +1408,8 @@ int mapThreadToCPUList(pthread_t tid, const std::set<int>& cpus)
   return pthread_setaffinity_np(tid,
                                 sizeof(cpuset),
                                 &cpuset);
-#endif /* HAVE_PTHREAD_SETAFFINITY_NP */
+#  endif
+#else
   return ENOSYS;
+#endif /* HAVE_PTHREAD_SETAFFINITY_NP */
 }
